// Copyright 2009, Gabriele Facciolo <gabriele.facciolo@upf.edu>
// Copyright 2009, Pablo Arias <pablo.arias@upf.edu>
// All rights reserved.

#include <algorithm>
#include <vector>
#include <iostream>
#include <math.h>
#include <stdlib.h>
#include "PointGrid.h"
#include "dimage.h"
#include "assert.h"
#include "smartparameter.h"

using namespace std;

#ifndef MAXFLOAT
#define MAXFLOAT 1000000000000.0
#endif
#define inf MAXFLOAT

#define max(a,b) (((a)>(b))?(a):(b))

SMART_PARAMETER(TYPE_A,0)
SMART_PARAMETER(TYPE_B,0)
SMART_PARAMETER(TYPE_AB,0)
SMART_PARAMETER(NOCENTER,1)
SMART_PARAMETER(NONORMALIZE,0)
SMART_PARAMETER(NONORMALIZEDUPDATE,0)

SMART_PARAMETER(MEDIANS,0)
SMART_PARAMETER(WINSIZE,500)
SMART_PARAMETER(WINDOWSTD,7)
SMART_PARAMETER(NEIGHSIZE,49)
SMART_PARAMETER(PATCH_RADIUS,0)
SMART_PARAMETER(DRYRUN,0)

double TAU = 0;

void printPoint (Point &a) {
	cerr  << a.x << " " << a.y << " " << a.value << endl;
}

double L2dist(Point &a, Point &b) {
	return ((a.x - b.x)*(a.x - b.x) + (a.y - b.y)*(a.y - b.y) );
}

// GLOBAL G function
double CMP_WINDOW_STD = 7;
double g_TABULATED[20000];
double gfun(double h2) {
	return exp(-h2/(CMP_WINDOW_STD*CMP_WINDOW_STD));
}
void init_g_TABULATED (void) {
	for(int i=0;i<20000;i++) { g_TABULATED[i] = gfun(i/8); }
}
double G(double h2) {
	return( (int)abs(h2*8) < 19999 ?  g_TABULATED[(int)abs(h2*8)]: g_TABULATED[19999]);
}
/*double G(double h2) {
	return exp(-h2/(CMP_WINDOW_STD*CMP_WINDOW_STD));
}*/



void center_Neighborhood( PointList &pl, Point &p) {
	for(int i = 0;  i< pl.size(); i++) { 
		pl[i].x -= p.x;
		pl[i].y -= p.y;
	}
}

void remove_Neighborhood_center( PointList &pl, Point &p) {
   PointList::iterator i;
   for(i = pl.begin();  i!= pl.end(); ++i) {
      if(((*i).x ==p.x ) && ((*i).y == p.y)) { i= pl.erase(i);}
   }
}


int WINDOW_CACHE=1;  // ACTIVATE ONLY IF YOU ARE GOING TO ITERATE
PointList **wind;

void getNearestNeighbours_center_cache(Dimage interp, PointGrid *grid, Point &m, int Nsize, PointList &refPL) { 
	if (WINDOW_CACHE) {
		int l = (int) m.x + (int)m.y*interp->ncol;
		if( wind[l]->size() == 0 ) { // if is not in the cache then compute it
			if( PATCH_RADIUS() == 0)
				*(wind[l]) = grid->getNearestNeighbours(m.x,m.y,Nsize); 
			else 
				*(wind[l]) = grid->getNearestNeighbours(m.x,m.y,Nsize, PATCH_RADIUS()); 
			center_Neighborhood( *(wind[l]),m);
		}
		refPL = *(wind[l]);	// the extract it
	} else { 
		if( PATCH_RADIUS() == 0)
			center_Neighborhood(  refPL = grid->getNearestNeighbours(m.x,m.y,Nsize), m);
		else
			center_Neighborhood(  refPL = grid->getNearestNeighbours(m.x,m.y,Nsize, PATCH_RADIUS()), m);
	}
}


/*

    vector<int> the_vector;
    vector<int>::iterator the_iterator;
 
    for( int i=0; i < 10; i++ ) the_vector.push_back(i);
    int total = 0;

    the_iterator = the_vector.begin();
    while( the_iterator != the_vector.end() ) {
      total += *the_iterator;
      ++the_iterator;
    }
 
    cout << "Total=" << total << endl;

*/

// receives the center, and two lists containing the relative coordinates of the samples and the image to be compared
double compute_dist_two( Point &m,  PointList &lhs, Point &P, PointList &rhs, Dimage im ) {
	double v=0, d=0 , n=0, g=0;
    vector<Point>::iterator it ;

    it = lhs.begin();
    while( it != lhs.end() ) {
		//		Check if the point is inside the image
		int X =(int) (m.x + it->x);
		int Y =(int) (m.y + it->y);

		if ((X>=0) && (Y>=0) && (Y<im->nrow) && (X<im->ncol) )  {
			// compute the index in the image of the current point
			int idx = X+ Y * im->ncol;
			if( MEDIANS() ) { 
				v = abs(it->value - im->gray[ idx ]);
			} else {
				v = (it->value - im->gray[ idx ]); 
				v*=v;
			}
			// compute the distance to the center
			//g = 1/ sqrt( 0.001+ lhs[i].x*lhs[i].x + lhs[i].y*lhs[i].y	);
			g = G(it->x*it->x + it->y*it->y);
			d += g*v;
			n += g;
		}
      ++it; 
   }

   it = rhs.begin();
   while( it != rhs.end() ) {
		//		Check if the point is inside the image
		int X =(int) (P.x + it->x);
		int Y =(int) (P.y + it->y);

		if ((X>=0) && (Y>=0) && (Y<im->nrow) && (X<im->ncol) )  {
			// compute the index in the image of the current point
			int idx = X+ Y * im->ncol;
			if( MEDIANS() ) { 
				v = abs(it->value - im->gray[ idx ]);
			} else {
				v = (it->value - im->gray[ idx ]); 
				v*=v;
			}
			// compute the distance to the center
			//g = 1/ sqrt( 0.001+ rhs[i].x*rhs[i].x + rhs[i].y*rhs[i].y	);
			g = G(it->x*it->x + it->y*it->y);
			d += g*v;
			n += g;
		}
      ++it; 
	}
	if(NONORMALIZE()) return(d);
	else return (n?d/n:0);
}

// receives the center, the relative coordinates of the samples and the image to be compared
double compute_dist( Point &m,  PointList &lhs, Dimage im ) {
	double v=0, d=0 , n=0, g=0;
	for(int i = 0;  i< lhs.size() ; i++) { 
		//		Check if the point is inside the image
		if (((m.x+lhs[i].x)>=0) && ((m.y+lhs[i].y)>=0) && ((m.y+lhs[i].y)<im->nrow) && ((m.x+lhs[i].x)<im->ncol) )  {
			// compute the index in the image of the current point
			int idx = (int) (m.x + lhs[i].x)+ (int)(lhs[i].y +m.y) * im->ncol;
			if( MEDIANS() ) { 
				v = abs(lhs[i].value - im->gray[ idx ]);
			} else {
				v = (lhs[i].value - im->gray[ idx ])*(lhs[i].value - im->gray[ idx ]);
			}
			// compute the distance to the center
			//g = 1/ sqrt( 0.001+ lhs[i].x*lhs[i].x + lhs[i].y*lhs[i].y	);
			g = G(lhs[i].x*lhs[i].x + lhs[i].y*lhs[i].y);
			d += g*v;
			n += g;
		}
	}
	if(NONORMALIZE()) return(d);
	else return (n?d/n:0);
}


inline void SWAPp(Point *P , Point *Q) { 
		Point t = *P;
		*P=*Q;
		*Q=t;
}
void addPT(PointList *p, Point newPT, int MAXF) {

	int REORDER =0;

	/* securuty *
	if(p->size()) {
		if((*p)[0].dist2 > (*p)[p->size()-1].dist2)  {	
			printf("%g %g\n", (*p)[0].dist2 , (*p)[p->size()-1].dist2) ;
		}
	}*/
//	if (p->size() && ((*p)[0].dist2 > (*p)[p->size()-1].dist2) ) printf( "%f %f\n", (*p)[0].dist2 , (*p)[p->size()-1].dist2);
	assert((p->size() && ((*p)[0].dist2 > (*p)[p->size()-1].dist2) ) == 0 );

	int n=p->size();
	if (n<MAXF) {
		p->push_back(newPT);
		REORDER =1;
	}
	else if ((*p)[MAXF-1].dist2 > newPT.dist2) {
		(*p)[MAXF-1] = newPT;
		REORDER =1;
	}

	if(REORDER) {
		int i=p->size()-1;
		while (i>=1 && (*p)[i].dist2 < (*p)[i-1].dist2) {
			SWAPp(&((*p)[i]), &((*p)[i-1]) );
			i--;
		}
	}
}
/* ONLY USEFUL FOR STORING THE MINIMUMS */
void addPTmindist2(PointList *p, Point newPT) {
	assert((p->size() == 0 || ((*p)[0].dist2 == (*p)[p->size()-1].dist2)));

	if( p->size() == 0 || newPT.dist2 == (*p)[0].dist2 ) { 
		p->push_back(newPT);
		return;
	}
	if( newPT.dist2 > (*p)[0].dist2 ) {
		return;
	}
	if( newPT.dist2 < (*p)[0].dist2 ) { 
		p->clear();
		p->push_back(newPT);
		return;
	}
}

void MEDIAN( PointList* bucket, double value, double w, double a ) {
		Point pt;
		pt.value = value;
		pt.dist2 = w;
		pt.g= a;
		bucket->push_back(pt);
}


void MEANlimB( PointList* bucket, double value, double dist2, double a , int indicated) {
	if( bucket->size() == 0) {
		Point pt;
		pt.value = value*a;
		pt.dist2 = dist2;
		pt.g= a;
		pt.index = (indicated?1:0);
		bucket->push_back(pt);
	} else {
		if( (*bucket)[0].index==0) {
			if(indicated==0) {
				if ( (*bucket)[0].dist2 ==  dist2 ) {
					(*bucket)[0].value += value*a;
					(*bucket)[0].g += a;
				} else if ((*bucket)[0].dist2 >  dist2 ) { 
					(*bucket)[0].value = value*a;
					(*bucket)[0].g = a;
					(*bucket)[0].dist2 = dist2;
				}
			} else { 
				(*bucket)[0].index = 1;
				(*bucket)[0].value = value*a;
				(*bucket)[0].g = a;
				(*bucket)[0].dist2 = dist2;
			}
		} else {
			if(indicated==1) {
				(*bucket)[0].value += value*a;
				(*bucket)[0].g += a;
				if ( (*bucket)[0].dist2 >  dist2 ) {
					(*bucket)[0].dist2 = dist2;
				}
			}
		}
	}
}


void MEANlim( PointList* bucket, double value, double dist2, double a ) {
	if( bucket->size() == 0) {
		Point pt;
		pt.value = value*a;
		pt.dist2 = dist2;
		pt.g= a;
		bucket->push_back(pt);
	} else {
		(*bucket)[0].value += value*a;
		(*bucket)[0].g += a;
		if ( (*bucket)[0].dist2 >  dist2 ) {
			(*bucket)[0].dist2 = dist2;
		}
	} 
}
void MEAN( PointList* bucket, double value, double w, double a ) {
		if( bucket->size() == 0) {
			Point pt;
			pt.value = value*w*a;
			pt.g = w*a;
			pt.dist2 = w;
			bucket->push_back(pt);
		} else {
			(*bucket)[0].value += value*w*a;
			(*bucket)[0].g += w*a;
			(*bucket)[0].dist2 = w;
		}
}

/* TYPE_A IMPLEMENTS the algorithm with reception of information to the center patch.  a(x') */
/* TYPE_B IMPLEMENTS the algorithm with transmission of information from the center patch.  a(x) */
/* TYPE_AB IMPLEMENTS the algorithm with transmission in both directions .  a(x) U a(x')*/
/* TYPE_O IMPLEMENTS the algorithm with transmission in both directions but using just  a(x)*/
#include "algorithmO.h"
#include "algorithmAB.h"
#include "algorithmA.h"
#include "algorithmB.h"



bool comparePointDist2(Point a, Point b)  {
	if(a.dist2 < b.dist2) return true;
	else	return false;
}

double weightedMEAN(PointList *pl, double *mass, double previous_value)  {
	double v= 0;
	*mass=0;
	if (pl->size() == 0) { 
		printf("empty bucket ID:%ld\n",pl->size());
	}	
	for(int i=0;i<pl->size();i++) {
		v += (*pl)[i].value ;
		*mass += (*pl)[i].g;
	}
	if(*mass != 0) return (v/ *mass);
	else return previous_value;
}

double weightedMean (PointList *pl, double *mass)  {
	double v= 0;
	*mass=0;

	// first sort all the candidates  by distance
	std::sort ( pl->begin(), pl->end() , comparePointDist2) ;
	assert( (*pl)[0].dist2  <= (*pl)[pl->size()-1].dist2 ) ;
	int NMIN=1;
	double dmin = (*pl)[0].dist2;

	while( NMIN < pl->size() && dmin == (*pl)[NMIN].dist2 ) NMIN++;

	for(int i=0;i<NMIN;i++) {
		//double weight = exp(-(*pl)[i].dist2 / 200) * (*pl)[i].a;
		double weight = (*pl)[i].g;
		v += (*pl)[i].value * weight;
		*mass +=weight;
	}
	v /= *mass;
	//if(NMIN>1) printf("%d %.2f %.2f %.2f %.2f\t",NMIN , dmin, *mass, (*pl)[0].value, v);
	//	*mass/=pl->size();
	return (v);
}

double weightedMeanOLD (PointList *pl, double *mass)  {
	double v= 0;
	*mass=0;
	for(int i=0;i<pl->size();i++) {
		//double weight = exp(-(*pl)[i].dist2 / 200) * (*pl)[i].a;
		double weight = (*pl)[i].dist2 * (*pl)[i].g;
		v += (*pl)[i].value * weight;
		*mass +=weight;
	}
	v /= *mass;
	//	*mass/=pl->size();
	return (v);
}

bool comparePoint(Point a, Point b)  {
	if(a.value < b.value) return true;
	else	return false;
}

double weightedMedian (PointList *pl, double * mass)  {
	double tot=0;
	for(int i=0;i<pl->size();i++) {
		double weight = (*pl)[i].dist2 * (*pl)[i].g;
		(*pl)[i].dist2 = weight;
		tot += (*pl)[i].dist2;
	}

	std::sort ( pl->begin(), pl->end() , comparePoint) ;

	double acc=0;
	for(int i=0;i<pl->size();i++) {
		acc += (*pl)[i].dist2;

		*mass =  (*pl)[i].dist2;
		if (acc>=tot/2) return ((*pl)[i].value);
		//		printf("%.2f ", (*pl)[i].value);
	}
// NEW
	return ((*pl)[pl->size()-1].value);
	//	printf("\n");
}




	extern "C" 
void NLinject_discrete(long numpix, double *DX, double *DY, double *U, long newPix, double *OUTDX, double *OUTDY, double *OUTU, Dimage interp, Dimage dist, double param_tau, int NLmeans)
{
	//	int Nneigh = 25, Nwindows = 50;
	//		int Nneigh = 25, Nwindows = 100;
	int Nneigh = 49, Nwindows = 500;
	//		int Nneigh = 81, Nwindows = 360;
	int N=50;

	TAU = param_tau;

	CMP_WINDOW_STD = (double) WINDOWSTD();
	Nneigh = (int) NEIGHSIZE();
	Nwindows = (int) WINSIZE();
	cout<< "Comparison window std: " << CMP_WINDOW_STD << "\tNeighbouring pixels in a window: "<< Nneigh << "\tTotal considered windows (per pixel): "<< Nwindows << "\tTAU: "<< TAU <<endl; cerr.flush();

	if(MEDIANS()) {cout<< "MEDIANS limit case "<<endl; cerr.flush();}
	if(TYPE_A())  {cout<< "TYPE_A "<<endl; cerr.flush();}
	if(TYPE_B())  {cout<< "TYPE_B "<<endl; cerr.flush();}
	if(TYPE_AB())  {cout<< "TYPE_AB "<<endl; cerr.flush();}
	if(!TYPE_A() && !TYPE_B() && !TYPE_AB()) {cout<< "TYPE_O"<<endl; cerr.flush();}
	if(NOCENTER()) {cout<< "NOCENTER"<<endl; cerr.flush();}
	if(NONORMALIZE()) {cout<< "NONORMALIZE"<<endl; cerr.flush();}
	if(NONORMALIZEDUPDATE()) {cout<< "NONORMALIZEDUPDATE"<<endl; cerr.flush();}


	init_g_TABULATED();

	// construct the GRID: 
	cout<< "Building search data structure. "<<endl; cerr.flush();
	PointGrid *grid = new PointGrid(DX,DY,U,numpix,0,0);

	// construct the buckets
	PointList** BIGOUT = new PointList*[interp->ncol*interp->nrow];
	for(int i=0; i<interp->ncol*interp->nrow; i++) BIGOUT[i] = new PointList();

	// prepare the cache of comparison windows
	if( WINDOW_CACHE) {
		wind = new PointList*[newPix];
		for (int l=0; l<newPix;l++) { wind[l] = new PointList; }
	}




	if(!DRYRUN())
	for (int l=0; l<newPix;l++) {
		Point m;
		m.x = OUTDX[l]; m.y = OUTDY[l]; m.value = OUTU[l];

		double weight;
		if ( TAU )  { 
			if ( TYPE_AB() )
				m = compute_bestmatch_at_pointAB( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
			else if( TYPE_A() )
				m = compute_bestmatch_at_pointA( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
			else if( TYPE_B() )
				m = compute_bestmatch_at_pointB( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
			else
				m = compute_bestmatch_at_pointO( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
		} else {
			if ( TYPE_AB() )
				m = compute_bestmatch_at_pointlimAB( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
			else if( TYPE_A() )
					m = compute_bestmatch_at_pointAlim( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
			else if( TYPE_B() )
					m = compute_bestmatch_at_pointBlim( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
			else 
					m = compute_bestmatch_at_pointOlim( grid, m, Nneigh, Nwindows, interp, &weight, BIGOUT,N); 
		}

		//dist->gray[(int) (m.x + m.y *interp->ncol)] = weight;

		if(!DRYRUN()){
			OUTDX[l] = m.x; OUTDY[l] = m.y; OUTU[l] = m.value; 
		}		
		if (! (l%interp->ncol)) printf("."); fflush(NULL);
	}

	printf("generate image\n");
	for (int y=0; y<interp->nrow;y++) {
		for (int x=0; x<interp->ncol;x++) {
			double v, mass; int l=y*interp->ncol + x;

			if(!DRYRUN()) {
				v = weightedMEAN(BIGOUT[l],&mass, interp->gray[l]);
				if(MEDIANS()) v = weightedMedian(BIGOUT[l],&mass);
				dist->gray[l] = mass;
				OUTDX[l] = x; OUTDY[l] = y; OUTU[l] = v; dist->gray[l] = mass;
			}
			else {
				dist->gray[l] = (OUTU[l]==-1000000?0:OUTU[l]);
				OUTDX[l] = x; OUTDY[l] = y; OUTU[l] = v; 
			}
		}
	}



	// display a neighborhood with weights 
	// get a neighborhood (near the corner of the image) 
	Point m;
	m.x=(int) 30;
	m.y=(int) 30;
	PointList PL;
	if(PATCH_RADIUS() == 0)
		PL = grid->getNearestNeighbours(m.x,m.y,Nneigh);
	else 
		PL = grid->getNearestNeighbours(m.x,m.y,Nneigh, PATCH_RADIUS());

	int maxDI =0;
	Point P;
	// determine the maximum distance to the center
	for (int i=0; i < PL.size(); i++) {
		P = PL[i];
		maxDI =  max( maxDI, (int) ceil( sqrt(L2dist(m,PL[i])))) ;
	}
	// plot the gaussian disk
	if(!DRYRUN()) {
		for (int i=-maxDI ; i<maxDI; i++ ) 
			for (int j=-maxDI ; j<maxDI; j++ ) { 
				if(m.x + i > 0 && m.x + i < dist->ncol && m.y + j > 0 && m.y + j < dist->nrow) { 
					dist->gray[(int) (m.x + i + (m.y + j)* dist->ncol)] = gfun((i*i+j*j))*1;
				}
			}
		for (int i=0; i < PL.size(); i++) {
			Point P = PL[i];
			dist->gray[(int) (PL[i].x + PL[i].y * dist->ncol)]=0;
		}
	}else {
		for (int i=-maxDI ; i<maxDI+1; i++ ) 
			for (int j=-maxDI ; j<maxDI+1; j++ ) { 
				if(maxDI + i >= 0 && maxDI + i < dist->ncol && maxDI + j >= 0 && maxDI + j < dist->nrow) { 
					dist->gray[(int) (maxDI + i + (maxDI + j)* dist->ncol)] = gfun((i*i+j*j))*256;
				}
			}
	}




	/* clean */
	for(int i=0; i<interp->ncol*interp->nrow; i++) delete BIGOUT[i];
   delete[] BIGOUT;
	delete grid;
   if( WINDOW_CACHE )  {
		for (int l=0; l<newPix;l++)  delete wind[l];
      delete[] wind;
   }
         
}








