// Copyright 2009, Gabriele Facciolo <gabriele.facciolo@upf.edu>
// Copyright 2009, Pablo Arias <pablo.arias@upf.edu>
// All rights reserved.

#include <math.h>
#include <vector>
#include "ann_1.1.1/include/ANN/ANN.h"             // ANN declarations

using namespace std;

typedef struct { float x, y; float value; long index; float dist2; float g;} Point;

/*class Point {
public:
	double x, y; double value; long index; double dist2;
}*/

typedef vector <Point> PointList;
typedef struct {
	Point center;
	float radius;
	PointList neighbors;
	float w;
} PointNeighborhood;

//typedef vector <Point>*vectorOfPoints;

class PointGrid //Point Grid Class
{ 
	private: 
		ANNkd_tree*       kdTree;              // search structure

		ANNpointArray     dataPts;          	// data points
		double *values;								// value of points
		ANNidxArray       nnIdx;               // near neighbor indices
		ANNdistArray      dists;               // near neighbor distances
		ANNpoint       queryPt;          		// query point
		int k;								// number of neighobors to be searched

		int dim ;

		long numelements;

	public: 
		// Constructor
		PointGrid(double *DX, double *DY, double *U, long numpix, long nx, long ny) 
		{ 
			dim=2;
			// store the values locally 
			numelements = numpix;
			values = new double[numelements];
	   	for(int i=0; i<numelements; i++) { values[i] = U[i]; }
			dataPts = annAllocPts(numelements, dim);       // allocate data points
	   	for(int i=0; i<numelements; i++) { 
				dataPts[i][0] = DX[i]; 
				dataPts[i][1] = DY[i]; 
			}

			// create the structure
		   kdTree = new ANNkd_tree(dataPts, numelements, dim);

			// setup input/outpu array for queries
			k=1; // BY DEFAULT IS 1-NN
			nnIdx = new ANNidx[k];                 // allocate near neigh indices
			dists = new ANNdist[k];                // allocate near neighbor dists
			queryPt = annAllocPt(dim);             // allocate query point
		} 

		void changeK( int newk) {
			if (newk>k) {

				delete [] nnIdx;                   // clean things up
				delete [] dists;
				nnIdx = new ANNidx[newk];                 // allocate near neigh indices
				dists = new ANNdist[newk];                // allocate near neighbor dists
			}
			k=newk; 
		}

		// Return the nearest point to the location (x,y) 
		Point getNearestPoint(double x, double y){
			Point minp;

			queryPt[0] = x; queryPt[1] = y;
			kdTree->annkSearch( queryPt, k,  nnIdx, dists,  0);

			minp.x = dataPts[nnIdx[0]][0];
			minp.y = dataPts[nnIdx[0]][1];
			minp.value = values[nnIdx[0]];
			minp.index = nnIdx[0];

			return minp;
		}
		
		long getNumPoints() {
			return numelements;
		}


		Point getPoint (int index) {
				Point p;
				p.x = dataPts[index][0];
				p.y = dataPts[index][1];
				p.value = values[index];
				p.index = index;
				return p;
		}
		
			
		// Return a point list of the K-nearest points to the location (x,y) 
		// contained in the ball of radius \sqrt(dist2)
		PointList getNearestNeighbours(double x, double y, int knn, double radius2 ){
			PointList pl;
			Point p;

			this->changeK(knn);

			queryPt[0] = x; queryPt[1] = y;
      	kdTree->annkFRSearch( queryPt, radius2 , knn, nnIdx, dists, 0 ); 

			
			for (int i = 0; i < knn; i++) {
				if (nnIdx[i]>=0) {
					p.x = dataPts[nnIdx[i]][0];
					p.y = dataPts[nnIdx[i]][1];
					p.value = values[nnIdx[i]];
					p.index = nnIdx[i];
					//dists[i] = sqrt(dists[i]);       // unsquare distance
					pl.push_back(p);
				}
			}
			return pl;
		}


		// Return a point list of the K-nearest points to the location (x,y) 
		PointList getNearestNeighbours(double x, double y, int knn){
			PointList pl;
			Point p;

			this->changeK(knn);

			queryPt[0] = x; queryPt[1] = y;
			kdTree->annkSearch( queryPt, knn,  nnIdx, dists,  0);
			
			for (int i = 0; i < knn; i++) {
				if (nnIdx[i]>=0) {
					p.x = dataPts[nnIdx[i]][0];
					p.y = dataPts[nnIdx[i]][1];
					p.value = values[nnIdx[i]];
					p.index = nnIdx[i];
					//dists[i] = sqrt(dists[i]);       // unsquare distance
					pl.push_back(p);
				}
			}
			return pl;
		}

		~PointGrid() //destructor for the C++ Tutorial 
		{ 
			delete [] nnIdx;                   // clean things up
			delete [] dists;
			delete [] values;
			delete kdTree;
         annDeallocPt(queryPt);
         annDeallocPts(dataPts);
			annClose();
		} 

};
