// Copyright 2009, Gabriele Facciolo <gabriele.facciolo@upf.edu>
// Copyright 2009, Pablo Arias <pablo.arias@upf.edu>
// All rights reserved.


/* 
 *
 *
 * TYPE_A IMPLEMENTS the algorithm with reception of information to the center patch.  a(x') 
 *
 *
 *
 * */

Point compute_bestmatch_at_pointA( PointGrid *grid, Point m, int Nsize, int WindSize, Dimage interp, double *weight, PointList** BIGOUT ,int N) {
	PointList searchNeigh,TP, putPL;
	double MINdist2 = -inf;
	assert(N< WindSize);

	// ATTENTION! this makes consider all the point
	//TODO	
	//N=WindSize;


	// GETS THE POINTS IN THE COMPARISON WINDOW
	searchNeigh =  grid->getNearestNeighbours(m.x,m.y,WindSize);  
	if(NOCENTER()) remove_Neighborhood_center(searchNeigh, m);

	// FOR EACH POINT IN searchNeigh
	for (int i=0; i < searchNeigh.size(); i++) {
		PointList refPL2;
		double dist2;
		Point P = searchNeigh[i];
		// GETS THE NEIGHBORHOOD OF THE CURRENT POINT (P) 
		getNearestNeighbours_center_cache(interp, grid, P, Nsize, refPL2);
		// COMPARE refPL WITH THE IMAGE
		dist2 = compute_dist( m, refPL2, interp) ;  // computes the mean error of all the pixels in the list against the interpolated image (reference)
		P.dist2 = dist2;
		addPT(&TP,P,N); // stores the N Points with shortest dist2
	}

	// compute the normalization constant
	double Wnorm = 0;
	for(int t=0;t<N;t++) {
		Point P = TP[t];
		Wnorm += exp(-P.dist2 / TAU); // COMPUTE w
	}
	//if(Wnorm==0) printf("W norm is 0\n");

	// accumulate the N neighborhoods
	for(int t=0;t<N;t++) {
		Point P = TP[t];

		// GET THE REMOTE NEIGHBORHOOD
		getNearestNeighbours_center_cache(interp, grid, P, Nsize, putPL);
		double mass = 0;  // compute the total mass
		for (int j=0; j < putPL.size(); j++) { 
			mass+=G(putPL[j].x*putPL[j].x + putPL[j].y*putPL[j].y); 
		}
		if(NONORMALIZE()) mass=1;

		// accumulate the pixels (all of them since it is a weighted mean)
		// MEAN case
		double w= exp(-P.dist2 / TAU); // COMPUTE w

		for (int j=0; j < putPL.size(); j++) {
			Point pact = putPL[j];
			int newx = (int) (m.x + pact.x );
			int newy = (int) (m.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Wnorm?w/Wnorm:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
				else
				{ MEAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Wnorm?w/Wnorm:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
			}
		}
	}

	// returns the minimal distance as well as the point
	*weight = TP[0].dist2;
	return (TP[0]);
}





/* 
 *
 *
 * TYPE_A IMPLEMENTS the algorithm with reception of information to the center patch.  a(x') 
 *
 *              IN THE LIMIT CASE (\tau \to 0)
 *
 * */



Point compute_bestmatch_at_pointAlim( PointGrid *grid, Point m, int Nsize, int WindSize, Dimage interp, double *weight, PointList** BIGOUT ,int N) {
	PointList TP, refPL ,searchNeigh,putPL;
	double MINdist2 = inf;
	Point Pmin;


	// GETS THE POINTS IN THE COMPARISON WINDOW
	searchNeigh =  grid->getNearestNeighbours(m.x,m.y,WindSize);  
	if(NOCENTER()) remove_Neighborhood_center(searchNeigh, m);

	// FOR EACH POINT IN searchNeigh
	for (int i=0; i < searchNeigh.size(); i++) {
		double dist2;
		Point P = searchNeigh[i];
		// GETS THE NEIGHBORHOOD OF THE CURRENT POINT (P) 
		getNearestNeighbours_center_cache(interp, grid, P, Nsize, refPL);
		dist2 = compute_dist( m, refPL, interp) ;  // computes the mean error of all the pixels in the list against the interpolated image (reference)
		P.dist2 = dist2;
		addPTmindist2(&TP,P); // stores the N Points with shortest dist2
	}

	// TODO: FIND OUT IF THIS IS CORRECT
	// SET THE Wnorm AS THE SIZE OF THE SET OF MINIMUM PIXELS
	double Wnorm = TP.size();
	if(TP.size()>1) printf("it happens %ld\n", TP.size());

	for(int t=0;t<TP.size();t++) {
		Point Pmin = TP[t];

		// GET THE EXTRACTED NEIGHBORHOOD
		getNearestNeighbours_center_cache(interp, grid, Pmin, Nsize, putPL);

		double mass = 0;  // compute the total mass
		for (int j=0; j < putPL.size(); j++) { 
			mass+=G(putPL[j].x*putPL[j].x + putPL[j].y*putPL[j].y); 
		}
		if(NONORMALIZE()) mass=1;

		// accumulate the pixels (all of them since it is a weighted mean)
		// MEAN case
		//	double w= exp(-Pmin.dist2 / TAU); // COMPUTE w

		for (int j=0; j < putPL.size(); j++) {
			Point pact = putPL[j];
			int newx = (int) (m.x + pact.x );
			int newy = (int) (m.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , 1 , G(pact.x*pact.x + pact.y*pact.y)/mass); }
				else 
				{ MEANlim( BIGOUT[ newx + newy * interp->ncol ], pact.value , Pmin.dist2 , G(pact.x*pact.x + pact.y*pact.y)/mass);}

			}
		}
	}
	// returns the minimal distance as well as the point
	*weight = TP[0].dist2;
	return (TP[0]);
}
