// Copyright 2009, Gabriele Facciolo <gabriele.facciolo@upf.edu>
// Copyright 2009, Pablo Arias <pablo.arias@upf.edu>
// All rights reserved.

/* 
 *
 *
 * TYPE_B IMPLEMENTS the algorithm with transmission of information from the center patch.  a(x) 
 *
 *
 *
 * */



Point compute_bestmatch_at_pointB( PointGrid *grid, Point m, int Nsize, int WindSize, Dimage interp, double *weight, PointList** BIGOUT ,int N) {
	PointList refPL ,searchNeigh,TP, putPL;
	double MINdist2 = -inf;
	assert(N< WindSize);

	// ATTENTION! this makes consider all the point
	//TODO	
	//N=WindSize;

	// GETS THE NEIGHBORHOOD OF THE CURRENT POINT (m)   x'
	getNearestNeighbours_center_cache(interp, grid, m, Nsize, refPL);

	// GETS THE POINTS IN THE COMPARISON WINDOW
	searchNeigh =  grid->getNearestNeighbours(m.x,m.y,WindSize);  
	if(NOCENTER()) remove_Neighborhood_center(searchNeigh, m);

	// FOR EACH POINT IN searchNeigh
	for (int i=0; i < searchNeigh.size(); i++) {
		double dist2;
		Point P = searchNeigh[i];   // x 
		// COMPARE refPL WITH THE IMAGE
		dist2 = compute_dist( P, refPL, interp) ;  // computes the mean error of all the pixels in the list against the interpolated image (reference)
		P.dist2 = dist2;
		addPT(&TP,P,N); // stores the N Points with shortest dist2
		/* compute the normalization factor */
	}

	// compute the normalization constant
	double Wnorm = 0;
	for(int t=0;t<N;t++) {
		Point P = TP[t];
		Wnorm += exp(-P.dist2 / TAU); // COMPUTE w
	}
	//if(Wnorm==0) printf("W norm is 0\n");
	//
	double mass = 0;  // compute the total mass
	for (int j=0; j < refPL.size(); j++) { 
		mass+=G(refPL[j].x*refPL[j].x + refPL[j].y*refPL[j].y); 
	}
	if(NONORMALIZE()) mass=1;

	// transfer the neighborhood N times to each destination
	for(int t=0;t<N;t++) {
		Point P = TP[t];

		// accumulate the pixels (all of them since it is a weighted mean)
		// MEAN case
		double w= exp(-P.dist2 / TAU); // COMPUTE w

		for (int j=0; j < refPL.size(); j++) {
			Point pact = refPL[j];
			int newx = (int) (P.x + pact.x );
			int newy = (int) (P.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Wnorm?w/Wnorm:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
				else
				{ MEAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Wnorm?w/Wnorm:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
			}
		}
	}

	// returns the minimal distance as well as the point
	*weight = TP[0].dist2;
	return (TP[0]);
}


/* 
 *
 *
 * TYPE_B IMPLEMENTS the algorithm with transmission of information from the center patch.  a(x) 
 *
 *              IN THE LIMIT CASE (\tau \to 0)
 *
 * */

Point compute_bestmatch_at_pointBlim( PointGrid *grid, Point m, int Nsize, int WindSize, Dimage interp, double *weight, PointList** BIGOUT ,int N) {
	PointList refPL ,searchNeigh,TP ;
	assert(N< WindSize);

	// GETS THE NEIGHBORHOOD OF THE CURRENT POINT (m) 
	getNearestNeighbours_center_cache(interp, grid, m, Nsize, refPL);

	double mass = 0;  // compute the total mass
	for (int j=0; j < refPL.size(); j++) { 
		mass+=G(refPL[j].x*refPL[j].x + refPL[j].y*refPL[j].y); 
	}	
	if(NONORMALIZE()) mass=1;	

	// GETS THE POINTS IN THE COMPARISON WINDOW
	searchNeigh =  grid->getNearestNeighbours(m.x,m.y,WindSize);  
	if(NOCENTER()) remove_Neighborhood_center(searchNeigh, m);

	// FOR EACH POINT IN searchNeigh
	for (int i=0; i < searchNeigh.size(); i++) {
		double dist2;
		Point P = searchNeigh[i];
		// COMPARE refPL WITH THE IMAGE
		dist2 = compute_dist( P, refPL, interp) ;  // computes the mean error of all the pixels in the list against the interpolated image (reference)
		P.dist2 = dist2;
		addPTmindist2(&TP,P); // stores the N Points with shortest dist2


		// ALWAYS STORE SOME VALUE
		// accumulate the base value 
		// (in case of not having a contrinution of a nearest neighbour) 
		for (int j=0; j < refPL.size(); j++) {
			Point pact = refPL[j];
			int newx = (int) (P.x + pact.x );
			int newy = (int) (P.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				MEANlimB( BIGOUT[ newx + newy * interp->ncol ], pact.value , dist2 , G(pact.x*pact.x + pact.y*pact.y)/mass,0);
			}
		}


	}

	// transfer the neighborhood N times to each destination
	for(int t=0;t<TP.size();t++) {
		Point P = TP[t];

		// accumulate the pixels (all of them since it is a weighted mean)
		for (int j=0; j < refPL.size(); j++) {
			Point pact = refPL[j];
			int newx = (int) (P.x + pact.x );
			int newy = (int) (P.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , 1 , G(pact.x*pact.x + pact.y*pact.y)/mass); }
				else 
				{ MEANlimB( BIGOUT[ newx + newy * interp->ncol ], pact.value , P.dist2 , G(pact.x*pact.x + pact.y*pact.y)/mass,1);}
			}
		}
	}

	// returns the minimal distance as well as the point
	*weight = TP[0].dist2;
	return (TP[0]);
}

