// Copyright 2009, Gabriele Facciolo <gabriele.facciolo@upf.edu>
// Copyright 2009, Pablo Arias <pablo.arias@upf.edu>
// All rights reserved.

/* 
 *
 *
 * TYPE_O IMPLEMENTS the algorithm with transmission of information in both ways but with  a(x) 
 *
 *
 *
 * */


Point compute_bestmatch_at_pointO( PointGrid *grid, Point m, int Nsize, int WindSize, Dimage interp, double *weight, PointList** BIGOUT ,int N) {
	PointList refPL ,searchNeigh,TP, putPL;
	assert(N<= WindSize);
	// ATTENTION! this makes consider all the point
	//TODO	
	//N=WindSize;

	// GETS THE NEIGHBORHOOD OF THE CURRENT POINT (m) 
	getNearestNeighbours_center_cache(interp, grid, m, Nsize, refPL);

	// GETS THE POINTS IN THE COMPARISON WINDOW
	searchNeigh =  grid->getNearestNeighbours(m.x,m.y,WindSize);  
	if(NOCENTER()) remove_Neighborhood_center(searchNeigh, m);

	// FOR EACH POINT IN searchNeigh
	for (int i=0; i < searchNeigh.size(); i++) {
		double dist2;
		Point P = searchNeigh[i];
		// COMPARE refPL WITH THE IMAGE
		dist2 = compute_dist( P, refPL, interp) ;  // computes the mean error of all the pixels in the list against the interpolated image (reference)
		P.dist2 = dist2;
		addPT(&TP,P,N); // stores the N Points with shortest dist2
		/* compute the normalization factor */
	}

	// compute the normalization constant
	double Rho= 0;
	for(int t=0;t<N;t++) {
		Point P = TP[t];
		Rho += exp(-P.dist2 / TAU); // COMPUTE r
	}
	//if(Rho ==0) printf("Rho is 0\n");

	// accumulate the N neighborhoods
	for(int t=0;t<N;t++) {
		Point P = TP[t];

		// GET THE EXTRACTED NEIGHBORHOOD
		getNearestNeighbours_center_cache(interp, grid, P, Nsize, putPL);

		double mass = 0;  // compute the total mass
		for (int j=0; j < refPL.size(); j++) { 
			mass+=G(refPL[j].x*refPL[j].x + refPL[j].y*refPL[j].y); 
		}
		for (int j=0; j < putPL.size(); j++) { 
			mass+=G(putPL[j].x*putPL[j].x + putPL[j].y*putPL[j].y); 
		}
		if(NONORMALIZE() || NONORMALIZEDUPDATE() ) mass=1;	

		// accumulate the pixels (all of them since it is a weighted mean)
		// MEAN case
		double w= exp(-P.dist2 / TAU); // COMPUTE w

		for (int j=0; j < putPL.size(); j++) {
			Point pact = putPL[j];
			int newx = (int) (m.x + pact.x );
			int newy = (int) (m.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Rho?w/Rho:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
				else
				{ MEAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Rho?w/Rho:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
			}
		}
		for (int j=0; j < refPL.size(); j++) {
			Point pact = refPL[j];
			int newx = (int) (P.x + pact.x );
			int newy = (int) (P.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Rho?w/Rho:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
				else
				{ MEAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , Rho?w/Rho:0 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
			}
		}
	}

	// returns the minimal distance as well as the point
	*weight = TP[0].dist2;
	return (TP[0]);
}


/* 
 *
 *
 * TYPE_O IMPLEMENTS the algorithm with transmission of information in both ways but with a(x) 
 *
 *              IN THE LIMIT CASE (\tau \to 0)
 *
 * */

Point compute_bestmatch_at_pointOlim( PointGrid *grid, Point m, int Nsize, int WindSize, Dimage interp, double *weight, PointList** BIGOUT ,int N) {
	PointList refPL ,searchNeigh,TP, putPL;

	// GETS THE NEIGHBORHOOD OF THE CURRENT POINT (m) 
	getNearestNeighbours_center_cache(interp, grid, m, Nsize, refPL);

	// GETS THE POINTS IN THE COMPARISON WINDOW
	searchNeigh =  grid->getNearestNeighbours(m.x,m.y,WindSize);  
	if(NOCENTER()) remove_Neighborhood_center(searchNeigh, m);

	// FOR EACH POINT IN searchNeigh
	for (int i=0; i < searchNeigh.size(); i++) {
		double dist2;
		Point P = searchNeigh[i];
		// COMPARE refPL WITH THE IMAGE
//		PointList refPL2;		
//		getNearestNeighbours_center_cache(interp, grid, P, Nsize, refPL2);
//		dist2 = compute_dist_two( m, refPL2, P,refPL, interp) ; 

		dist2 = compute_dist( P, refPL, interp) ;  // computes the mean error of all the pixels in the list against the interpolated image (reference)
		P.dist2 = dist2;
		addPTmindist2(&TP,P); // stores the N Points with shortest dist2
	}

	// perform the transfers
	for(int t=0;t<TP.size();t++) {
		Point P = TP[t];

		// GET THE REMOTE NEIGHBORHOOD
		getNearestNeighbours_center_cache(interp, grid, P, Nsize, putPL);

		double mass = 0;  // compute the total mass
		for (int j=0; j < refPL.size(); j++) { 
			mass+=G(refPL[j].x*refPL[j].x + refPL[j].y*refPL[j].y); 
		}
		for (int j=0; j < putPL.size(); j++) { 
			mass+=G(putPL[j].x*putPL[j].x + putPL[j].y*putPL[j].y); 
		}
		if(NONORMALIZE() || NONORMALIZEDUPDATE() ) mass=1;	

		for (int j=0; j < putPL.size(); j++) {
			Point pact = putPL[j];
			int newx = (int) (m.x + pact.x );
			int newy = (int) (m.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , 1 , G(pact.x*pact.x + pact.y*pact.y)/mass); }
				else 
				{ MEANlim( BIGOUT[ newx + newy * interp->ncol ], pact.value , P.dist2 , G(pact.x*pact.x + pact.y*pact.y)/mass);}

			}
		}
		for (int j=0; j < refPL.size(); j++) {
			Point pact = refPL[j];
			int newx = (int) (P.x + pact.x );
			int newy = (int) (P.y + pact.y );
			if(newy>=0 && newx>=0 && newy < interp->nrow && newx < interp->ncol) {
				if (MEDIANS() ) 
				{ MEDIAN( BIGOUT[ newx + newy * interp->ncol ], pact.value , 1 , G(pact.x*pact.x + pact.y*pact.y)/mass); }
				else 
				{ MEANlim( BIGOUT[ newx + newy * interp->ncol ], pact.value , P.dist2 , G(pact.x*pact.x + pact.y*pact.y)/mass);}
			}
		}
	}

	// returns the minimal distance as well as the point
	*weight = TP[0].dist2;
	return (TP[0]);
}
